from typing import Optional


class MistralCommonException(Exception):
    message: str = "Internal server error"

    def __init__(
        self,
        message: Optional[str] = None,
    ) -> None:
        if message:
            self.message = message


class TokenizerException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class UnsupportedTokenizerFeatureException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidRequestException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidSystemPromptException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidMessageStructureException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidAssistantMessageException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidToolMessageException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidToolSchemaException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidUserMessageException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidFunctionCallException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidToolException(MistralCommonException):
    def __init__(self, message: str) -> None:
        super().__init__(message)



from pydantic import BaseModel, ConfigDict


class MistralBase(BaseModel):
    """
    Base class for all Mistral Pydantic models.
    """

    model_config = ConfigDict(extra="forbid", strict=True, validate_default=True, use_enum_values=True)



from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import Field
from typing_extensions import Annotated  # compatibility with 3.8

from mistral_common.base import MistralBase
from mistral_common.protocol.instruct.tool_calls import ToolCall


class Roles(str, Enum):
    system = "system"
    user = "user"
    assistant = "assistant"
    tool = "tool"


class ChunkTypes(str, Enum):
    text = "text"


class ContentChunk(MistralBase):
    type: ChunkTypes = ChunkTypes.text
    text: str


class BaseMessage(MistralBase):
    role: Literal[Roles.system, Roles.user, Roles.assistant, Roles.tool]


class UserMessage(BaseMessage):
    role: Literal[Roles.user] = Roles.user
    content: Union[str, List[ContentChunk]]


class SystemMessage(BaseMessage):
    role: Literal[Roles.system] = Roles.system
    content: Union[str, List[ContentChunk]]


class AssistantMessage(BaseMessage):
    role: Literal[Roles.assistant] = Roles.assistant
    content: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


class ToolMessage(BaseMessage):
    content: str
    role: Literal[Roles.tool] = Roles.tool
    tool_call_id: Optional[str] = None

    # Deprecated in V3 tokenization
    name: Optional[str] = None


ChatMessage = Annotated[Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage], Field(discriminator="role")]



import json
from typing import List, Optional, Sequence, Union

from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    ContentChunk,
    Roles,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.request import ChatCompletionRequest, ChatMessage
from mistral_common.protocol.instruct.tool_calls import FunctionCall, ToolCall
from mistral_common.tokens.instruct.request import InstructRequest


class InstructRequestNormalizer:
    """
    Takes a ChatCompletionRequest and normalizes it into an InstructRequest.

    The normalization process does several things such as:
    - Aggregate consecutive messages of the same role
    - Aggregate system prompts
    - Normalize json content
    - Normalize tool calls
    """

    def _normalize_json_content(self, content: Optional[str]) -> str:
        if content is None or len(content) == 0:
            return "{}"

        try:
            parsed_json = json.loads(content)
            normalized_content = json.dumps(parsed_json, ensure_ascii=False)
        except json.JSONDecodeError:
            normalized_content = content
        return normalized_content

    def _aggregate_content_chunks(self, content: Union[str, List[ContentChunk]], chunk_join_str: str = "\n\n") -> str:
        if isinstance(content, list):
            return chunk_join_str.join([chunk.text for chunk in content])
        else:
            return content

    def _aggregate_system_prompts(self, request: ChatCompletionRequest) -> Optional[str]:
        system_prompt: List[str] = []

        for message in request.messages:
            if message.role == Roles.system and message.content:
                system_prompt.append(self._aggregate_content_chunks(message.content))

        return "\n\n".join(system_prompt) if len(system_prompt) else None

    def _aggregate_tool_messages(self, messages: List[ChatMessage]) -> List[ToolMessage]:
        """
        We currently do not do any aggregation for tool messages, but we normalize the json content
        """
        tool_messages: List[ToolMessage] = []
        for message in messages:
            assert isinstance(message, ToolMessage), "Expected tool message"
            content = self._aggregate_content_chunks(message.content)
            normalized_content = self._normalize_json_content(content)
            tool_messages.append(
                ToolMessage(content=normalized_content, tool_call_id=message.tool_call_id, name=message.name)
            )

        return tool_messages

    def _normalize_tool_call(self, tool_call: ToolCall) -> ToolCall:
        normalized_function_aruments = self._normalize_json_content(tool_call.function.arguments)
        return ToolCall(
            function=FunctionCall(name=tool_call.function.name, arguments=normalized_function_aruments),
            id=tool_call.id,
        )

    def _aggregate_assistant_messages(self, messages: List[ChatMessage]) -> AssistantMessage:
        aggregated_content: List[str] = []
        tool_calls: List[ToolCall] = []
        for message in messages:
            assert isinstance(message, AssistantMessage), "Expected assistant message"
            if message.tool_calls is not None:
                for tool_call in message.tool_calls:
                    normalized_tool_call = self._normalize_tool_call(tool_call)
                    tool_calls.append(normalized_tool_call)
            elif message.content:
                aggregated_content.append(self._aggregate_content_chunks(message.content))

        return AssistantMessage(
            content="\n\n".join(aggregated_content) if len(aggregated_content) else None,
            tool_calls=tool_calls or None,
        )

    def _aggregate_user_messages(self, messages: List[ChatMessage]) -> UserMessage:
        aggregated_content: List[str] = []
        for message in messages:
            assert isinstance(message, UserMessage), "Expected user message"
            content = self._aggregate_content_chunks(message.content)
            if content:
                aggregated_content.append(content)

        aggregated_content_str = "\n\n".join(aggregated_content)
        return UserMessage(content=aggregated_content_str)

    def _aggregate_role(self, messages: List[ChatMessage], role: Optional[Roles]) -> Sequence[ChatMessage]:
        if role == Roles.tool:
            return self._aggregate_tool_messages(messages)
        elif role == Roles.assistant:
            return [self._aggregate_assistant_messages(messages)]
        elif role == Roles.user:
            return [self._aggregate_user_messages(messages)]
        else: # System messages are ignored
            return []

    def _aggregate_messages(self, request: ChatCompletionRequest) -> List[ChatMessage]:
        aggregated_messages: List[ChatMessage] = []
        messages_to_aggregate: List[ChatMessage] = []
        current_role: Optional[Roles] = None

        # Collect consecutive lists of messages with the same role
        for message in request.messages:
            if current_role != message.role:
                aggregated_messages.extend(self._aggregate_role(messages_to_aggregate, current_role))
                messages_to_aggregate.clear()

            current_role = message.role
            messages_to_aggregate.append(message)

        # Add the last set of messages
        aggregated_messages.extend(self._aggregate_role(messages_to_aggregate, current_role))

        # If the first message is not a user message, or we didnt aggregate
        # anything (all system messages) for example, add an empty user message
        if len(aggregated_messages) == 0 or aggregated_messages[0].role != Roles.user:
            aggregated_messages.insert(0, UserMessage(content=""))

        return aggregated_messages

    def from_chat_completion_request(self, request: ChatCompletionRequest) -> InstructRequest:
        system_prompt = self._aggregate_system_prompts(request)
        messages = self._aggregate_messages(request)

        return InstructRequest(messages=messages, system_prompt=system_prompt, available_tools=request.tools)



from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Protocol

from mistral_common.base import MistralBase
from mistral_common.tokens.instruct.request import InstructRequest


class SpecialTokens(str, Enum):
    bos = "<s>"
    eos = "</s>"
    begin_inst = "[INST]"
    end_inst = "[/INST]"
    begin_tools = "[AVAILABLE_TOOLS]"
    end_tools = "[/AVAILABLE_TOOLS]"
    begin_tool_results = "[TOOL_RESULTS]"
    end_tool_results = "[/TOOL_RESULTS]"
    tool_calls = "[TOOL_CALLS]"


class Tokenized(MistralBase):
    """
    A tokenized InstructRequest
    """

    tokens: List[int]
    text: Optional[str] = None


class Tokenizer(ABC):
    @property
    @abstractmethod
    def n_words(self) -> int:
        """Vocabulary size"""

    @abstractmethod
    def vocab(self) -> List[str]:
        """All tokens in the vocabulary as strings"""

    @property
    @abstractmethod
    def bos_id(self) -> int:
        """id of the Beginning of String token"""

    @property
    @abstractmethod
    def eos_id(self) -> int:
        """id of the End of String token"""

    @abstractmethod
    def encode(self, s: str, bos: bool, eos: bool) -> List[int]:
        """String to token ids"""

    @abstractmethod
    def decode(self, t: List[int]) -> str:
        """Token ids to string"""


class InstructTokenizer(Protocol):
    def encode_instruct(self, request: InstructRequest) -> Tokenized:
        """Instruct request to Tokenized object"""



from enum import Enum
from typing import List, Optional

from pydantic import Field

from mistral_common.base import MistralBase
from mistral_common.protocol.instruct.messages import ChatMessage
from mistral_common.protocol.instruct.tool_calls import Tool, ToolChoice


class ResponseFormats(str, Enum):
    text: str = "text"
    json: str = "json_object"


class ResponseFormat(MistralBase):
    type: ResponseFormats = ResponseFormats.text


class ChatCompletionRequest(MistralBase):
    model: Optional[str] = None
    messages: List[ChatMessage]
    response_format: ResponseFormat = Field(default_factory=ResponseFormat)
    tools: List[Tool] = Field(default_factory=list)
    tool_choice: ToolChoice = ToolChoice.auto
    temperature: float = Field(default=0.7, ge=0.0, le=1.0)
    top_p: float = Field(default=1.0, ge=0.0, le=1.0)
    max_tokens: Optional[int] = Field(default=None, ge=0)
    random_seed: Optional[int] = Field(default=None, ge=0)



from enum import Enum
from typing import Any, Dict

from mistral_common.base import MistralBase


class Function(MistralBase):
    name: str
    description: str = ""
    parameters: Dict[str, Any]


class ToolType(str, Enum):
    function = "function"


class ToolChoice(str, Enum):
    auto: str = "auto"
    none: str = "none"
    any: str = "any"


class Tool(MistralBase):
    type: ToolType = ToolType.function
    function: Function


class FunctionCall(MistralBase):
    name: str
    arguments: str


class ToolCall(MistralBase):
    id: str = "null"  # required for V3 tokenization
    type: ToolType = ToolType.function
    function: FunctionCall



import time
from enum import Enum
from typing import List, Optional

from pydantic import Field

from mistral_common.base import MistralBase
from mistral_common.protocol.base import UsageInfo
from mistral_common.protocol.instruct.tool_calls import ToolCall
from mistral_common.protocol.utils import random_uuid


class FinishReason(str, Enum):
    stop: str = "stop"
    length: str = "length"
    model_length: str = "model_length"
    error: str = "error"
    tool_call: str = "tool_calls"


class ChatCompletionTokenLogprobs(MistralBase):
    token: str
    logprob: float
    bytes: List[int]


class ChatCompletionResponseChoiceLogprobs(MistralBase):
    content: List[ChatCompletionTokenLogprobs]


class DeltaMessage(MistralBase):
    role: Optional[str] = None
    content: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


class ChatCompletionResponseChoice(MistralBase):
    index: int
    message: DeltaMessage
    finish_reason: Optional[FinishReason] = None
    logprobs: Optional[ChatCompletionResponseChoiceLogprobs] = None


class ChatCompletionResponse(MistralBase):
    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo


class ChatCompletionResponseStreamChoice(MistralBase):
    index: int
    delta: DeltaMessage
    finish_reason: Optional[FinishReason] = None
    logprobs: Optional[ChatCompletionResponseChoiceLogprobs] = None


class ChatCompletionStreamResponse(MistralBase):
    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion.chunk"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseStreamChoice]
    usage: Optional[UsageInfo] = None



import json
import logging
import os
from abc import abstractmethod
from functools import cached_property
from typing import Any, Dict, List, Optional, Set

from mistral_common.exceptions import TokenizerException
from mistral_common.protocol.instruct.messages import (
    AssistantMessage,
    ToolMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.tool_calls import Tool, ToolCall
from mistral_common.tokens.instruct.request import InstructRequest
from mistral_common.tokens.tokenizers.base import (
    InstructTokenizer,
    SpecialTokens,
    Tokenized,
    Tokenizer,
)
from sentencepiece import SentencePieceProcessor


class SentencePieceTokenizer(Tokenizer):
    def __init__(self, model_path: str) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        # reload tokenizer
        assert os.path.isfile(model_path), model_path
        self._model = SentencePieceProcessor(model_file=model_path)

        assert self._model.vocab_size() == self._model.get_piece_size()
        self._vocab = [self._model.id_to_piece(i) for i in range(self.n_words)]

        super().__init__()

    @property
    def n_words(self) -> int:
        return self._model.vocab_size()  # type: ignore

    def vocab(self) -> List[str]:
        return self._vocab

    @property
    def bos_id(self) -> int:
        return self._model.bos_id()  # type: ignore

    @property
    def eos_id(self) -> int:
        return self._model.eos_id()  # type: ignore

    @cached_property
    def _control_tokens(self) -> Set[int]:
        return {tok for tok in range(self.n_words) if self._model.IsControl(tok)}

    def encode(self, s: str, bos: bool, eos: bool) -> List[int]:
        assert isinstance(s, str)
        t: List[int] = self._model.encode(s)
        if bos:
            t = [self.bos_id, *t]
        if eos:
            t = [*t, self.eos_id]
        return t

    def decode(self, t: List[int]) -> str:
        return self._model.decode(t)  # type: ignore

    def id_to_piece(self, token_id: int) -> str:
        return self._model.id_to_piece(token_id)  # type: ignore

    def to_string(self, tokens: List[int]) -> str:
        """
        Converts tokens into a string for debugging purposes
        """
        text = ""
        curr_tokens: List[int] = []
        for tok in tokens:
            if tok in self._control_tokens:
                if curr_tokens:
                    text += "".join([self.id_to_piece(tok) for tok in curr_tokens])
                    curr_tokens = []

                text += self.id_to_piece(tok)

            else:
                curr_tokens.append(tok)

        if curr_tokens:
            text += "".join([self.id_to_piece(tok) for tok in curr_tokens])

        return text


class SentencePieceInstructTokenizer(InstructTokenizer):
    def __init__(self, model_path: str):
        self.tokenizer = SentencePieceTokenizer(model_path)

    def start(self) -> List[int]:
        return [self.tokenizer.bos_id]

    @abstractmethod
    def encode_user_message(
        self,
        message: UserMessage,
        available_tools: Optional[List[Tool]],
        is_last: bool,
        is_first: bool,
        system_prompt: Optional[str] = None,
    ) -> List[int]:
        ...

    @abstractmethod
    def encode_tool_message(self, message: ToolMessage, is_before_last_user_message: bool) -> List[int]:
        ...

    @abstractmethod
    def encode_assistant_message(self, message: AssistantMessage, is_before_last_user_message: bool) -> List[int]:
        ...

    def encode_instruct(self, request: InstructRequest) -> Tokenized:
        # init at bos
        tokens = self.start()
        # find last user message
        last_user_idx = -1
        first_user_idx = -1
        for i, msg in list(enumerate(request.messages)):
            if isinstance(msg, UserMessage):
                if first_user_idx == -1:
                    first_user_idx = i
                last_user_idx = i
        for msg_idx, msg in enumerate(request.messages):
            if isinstance(msg, UserMessage):
                new_tokens = self.encode_user_message(
                    msg,
                    request.available_tools,
                    msg_idx == last_user_idx,
                    msg_idx == first_user_idx,
                    system_prompt=request.system_prompt,
                )
            elif isinstance(msg, ToolMessage):
                new_tokens = self.encode_tool_message(msg, msg_idx < last_user_idx)
            elif isinstance(msg, AssistantMessage):
                new_tokens = self.encode_assistant_message(msg, msg_idx < last_user_idx)

            tokens.extend(new_tokens)

        return Tokenized(tokens=tokens, text=self.tokenizer.to_string(tokens))


class SentencePieceInstructTokenizerV1(SentencePieceInstructTokenizer):
    def encode_user_message(
        self,
        message: UserMessage,
        available_tools: Optional[List[Tool]],
        is_last: bool,
        is_first: bool,
        system_prompt: Optional[str] = None,
    ) -> List[int]:
        assert message.content is not None
        assert isinstance(message.content, str), "Message content must be nornmalized"
        content = ""
        if is_first and system_prompt:
            content = system_prompt + "\n\n" + message.content
        else:
            content = message.content

        message_txt = f"[INST] {content} [/INST]"
        curr_tokens = self.tokenizer.encode(message_txt, bos=False, eos=False)
        return curr_tokens

    def encode_tool_message(self, message: ToolMessage, is_before_last_user_message: bool) -> List[int]:
        raise TokenizerException("Tools not implemented for tokenizer V1")

    def encode_assistant_message(self, message: AssistantMessage, is_before_last_user_message: bool) -> List[int]:
        assert isinstance(message, AssistantMessage), message
        if message.tool_calls is not None and len(message.tool_calls) > 0:
            raise TokenizerException("Tools not implemented for tokenizer V1")
        elif message.content:
            curr_tokens = self.tokenizer.encode(message.content, bos=False, eos=False)
        else:
            raise TokenizerException(f"{message.content} // {message.tool_calls}")

        curr_tokens.append(self.tokenizer.eos_id)
        return curr_tokens


class SentencePieceInstructTokenizerV2(SentencePieceInstructTokenizer):
    def __init__(self, model_path: str):
        super().__init__(model_path)

        self.BEGIN_INST = self.get_control_token(SpecialTokens.begin_inst.value)
        self.END_INST = self.get_control_token(SpecialTokens.end_inst.value)
        self.BEGIN_AVAILABLE_TOOLS = self.get_control_token(SpecialTokens.begin_tools.value)
        self.END_AVAILABLE_TOOLS = self.get_control_token(SpecialTokens.end_tools.value)
        self.BEGIN_TOOL_RESULTS = self.get_control_token(SpecialTokens.begin_tool_results.value)
        self.END_TOOL_RESULTS = self.get_control_token(SpecialTokens.end_tool_results.value)
        self.TOOL_CALLS = self.get_control_token(SpecialTokens.tool_calls.value)

    def get_control_token(self, s: str) -> int:
        return self.tokenizer._model.piece_to_id(s)  # type: ignore

    def encode_user_message(
        self,
        message: UserMessage,
        available_tools: Optional[List[Tool]],
        is_last: bool,
        is_first: bool,
        system_prompt: Optional[str] = None,
    ) -> List[int]:
        assert message.content is not None
        assert isinstance(message.content, str), "Message content must be nornmalized"
        content = ""
        tools_tokens: List[int] = []
        if is_last and available_tools:
            tools = [tool.model_dump() for tool in available_tools]
            tools_json_tokens = self.tokenizer.encode(json.dumps(tools, ensure_ascii=False), bos=False, eos=False)
            tools_tokens = [
                self.BEGIN_AVAILABLE_TOOLS,
                *tools_json_tokens,
                self.END_AVAILABLE_TOOLS,
            ]

        if is_last and system_prompt:
            content = system_prompt + "\n\n" + message.content
        else:
            content = message.content

        curr_tokens = [
            *tools_tokens,
            self.BEGIN_INST,
            *self.tokenizer.encode(content, bos=False, eos=False),
            self.END_INST,
        ]
        return curr_tokens

    def _parse_json_content(self, content: str) -> Any:
        try:
            return json.loads(content)
        except json.JSONDecodeError:
            return content

    def _prepare_tool_result(self, tool_message: ToolMessage) -> Dict[str, Any]:
        """
        Bit of a hack due to the way tool results are tokenized
        """
        assert tool_message.content is not None, "Tool message content cannot be None"
        return {
            "name": tool_message.name,
            "content": self._parse_json_content(tool_message.content),
        }

    def encode_tool_message(self, message: ToolMessage, is_before_last_user_message: bool) -> List[int]:
        if is_before_last_user_message:
            # don't tokenize last tool response before last user msg
            return []

        # Currently only supports single tool results
        tool_result_str = json.dumps([self._prepare_tool_result(message)], ensure_ascii=False)
        curr_tokens = [
            self.BEGIN_TOOL_RESULTS,
            *self.tokenizer.encode(tool_result_str, bos=False, eos=False),
            self.END_TOOL_RESULTS,
        ]
        return curr_tokens

    def _prepare_function_call(self, tool_call: ToolCall) -> Dict[str, Any]:
        """
        Bit of a hack due to the way function calls are tokenized
        """
        return {
            "name": tool_call.function.name,
            "arguments": self._parse_json_content(tool_call.function.arguments),
        }

    def encode_assistant_message(self, message: AssistantMessage, is_before_last_user_message: bool) -> List[int]:
        if message.tool_calls is not None and len(message.tool_calls) > 0:
            if is_before_last_user_message:
                # don't tokenize tool call before last user message
                return []

            prepared_tool_calls = []
            for tool_call in message.tool_calls:
                prepared_tool_calls.append(self._prepare_function_call(tool_call))

            tool_call_str = json.dumps(prepared_tool_calls, ensure_ascii=False)
            curr_tokens = [
                self.TOOL_CALLS,
                *self.tokenizer.encode(tool_call_str, bos=False, eos=False),
            ]
        elif message.content:
            curr_tokens = self.tokenizer.encode(message.content, bos=False, eos=False)
        else:
            raise TokenizerException(f"Invalid assistant message: {message.content}")

        curr_tokens.append(self.tokenizer.eos_id)
        return curr_tokens


class SentencePieceInstructTokenizerV3(SentencePieceInstructTokenizerV2):
    """
    The only difference with V3 tokenizer is that it encodes the tool messages differently
    """

    def _prepare_function_call(self, tool_call: ToolCall) -> Dict[str, Any]:
        return {
            "name": tool_call.function.name,
            "arguments": self._parse_json_content(tool_call.function.arguments),
            "id": tool_call.id,
        }

    def _prepare_tool_result(self, tool_message: ToolMessage) -> Dict[str, Any]:
        assert tool_message.content is not None, "Tool message content cannot be None"
        return {
            "call_id": tool_message.tool_call_id,
            "content": self._parse_json_content(tool_message.content),
        }

    def encode_tool_message(self, message: ToolMessage, is_before_last_user_message: bool) -> List[int]:
        """
        Same as V2 but tools not wrapped in a list and history is tokenized also
        """
        tool_result_str = json.dumps(self._prepare_tool_result(message))
        curr_tokens = [
            self.BEGIN_TOOL_RESULTS,
            *self.tokenizer.encode(tool_result_str, bos=False, eos=False),
            self.END_TOOL_RESULTS,
        ]
        return curr_tokens

    def encode_assistant_message(self, message: AssistantMessage, is_before_last_user_message: bool) -> List[int]:
        """
        Same as V2 but always encode tool history
        """
        return super().encode_assistant_message(message, False)



from __future__ import annotations

from pathlib import Path
from typing import Type

from mistral_common.exceptions import (
    TokenizerException,
)
from mistral_common.protocol.instruct.request import ChatCompletionRequest
from mistral_common.protocol.instruct.validator import (
    MistralRequestValidator,
    MistralRequestValidatorV3,
    ValidationMode,
)
from mistral_common.tokens.instruct.normalize import InstructRequestNormalizer
from mistral_common.tokens.tokenizers.base import InstructTokenizer, Tokenized
from mistral_common.tokens.tokenizers.sentencepiece import (
    SentencePieceInstructTokenizerV1,
    SentencePieceInstructTokenizerV2,
    SentencePieceInstructTokenizerV3,
)


class MistralTokenizer:
    def __init__(
        self,
        instruct_tokenizer: InstructTokenizer,
        validator: Type[MistralRequestValidator] = MistralRequestValidator,
        request_normalizer: Type[InstructRequestNormalizer] = InstructRequestNormalizer,
        mode: ValidationMode = ValidationMode.test,
    ):
        self._chat_completion_request_validator = validator(mode)
        self._instruct_request_normalizer = request_normalizer()
        self.instruct_tokenizer = instruct_tokenizer


    @classmethod
    def _data_path(cls) -> Path:
        return Path(__file__).parents[2] / "data"

    @classmethod
    def v1(cls) -> MistralTokenizer:
        """open-mistral-7B // open-mixtral-8x7B // mistral-embed"""
        return cls.from_file(str(cls._data_path() / "tokenizer.model.v1"), mode=ValidationMode.test)

    @classmethod
    def v2(cls) -> MistralTokenizer:
        """mistral-small // mistral-large"""
        return cls.from_file(
            str(cls._data_path() / "mistral_instruct_tokenizer_240216.model.v2"), mode=ValidationMode.test
        )

    @classmethod
    def v3(cls) -> MistralTokenizer:
        """open-mixtral-8x22B"""
        return cls.from_file(
            str(cls._data_path() / "mistral_instruct_tokenizer_240216.model.v3"), mode=ValidationMode.test
        )

    @classmethod
    def from_model(cls, model: str) -> MistralTokenizer:
        model_name_to_tokenizer_cls = {
            "open-mistral-7b": MistralTokenizer.v1,
            "open-mixtral-8x7b": MistralTokenizer.v1,
            "mistral-embed": MistralTokenizer.v1,
            "mistral-small": MistralTokenizer.v2,
            "mistral-large": MistralTokenizer.v2,
            "open-mixtral-8x22b": MistralTokenizer.v3,
        }

        # Prefix search the model name mapping
        for model_name, tokenizer_cls in model_name_to_tokenizer_cls.items():
            if model_name in model:
                return tokenizer_cls()

        raise TokenizerException(f"Unrecognized model: {model}")

    @classmethod
    def from_file(cls, tokenizer_filename: str, mode: ValidationMode = ValidationMode.test) -> MistralTokenizer:
        """
        Depending on which model we are loading, tokenization and validation might be different. 💩
        """
        if tokenizer_filename.endswith(".model.v1"):
            return cls(
                SentencePieceInstructTokenizerV1(tokenizer_filename),
                validator=MistralRequestValidator,
                request_normalizer=InstructRequestNormalizer,
                mode=mode,
            )
        elif tokenizer_filename.endswith(".model.v2"):
            return cls(
                SentencePieceInstructTokenizerV2(tokenizer_filename),
                validator=MistralRequestValidator,
                request_normalizer=InstructRequestNormalizer,
                mode=mode,
            )
        elif tokenizer_filename.endswith(".model.v3"):
            return cls(
                SentencePieceInstructTokenizerV3(tokenizer_filename),
                validator=MistralRequestValidatorV3,
                request_normalizer=InstructRequestNormalizer,
                mode=mode,
            )
        elif tokenizer_filename.endswith(".model"):
            return cls(
                SentencePieceInstructTokenizerV1(tokenizer_filename),
                validator=MistralRequestValidator,
                request_normalizer=InstructRequestNormalizer,
                mode=mode,
            )
        else:
            raise TokenizerException(f"Unrecognized tokenizer filename: {tokenizer_filename}")

    def encode_chat_completion(self, request: ChatCompletionRequest) -> Tokenized:
        validated_request = self._chat_completion_request_validator.validate_request(request)
        instruct_request = self._instruct_request_normalizer.from_chat_completion_request(validated_request)
        return self.instruct_tokenizer.encode_instruct(instruct_request)



from typing import List, Optional

from mistral_common.base import MistralBase
from mistral_common.protocol.instruct.messages import ChatMessage
from mistral_common.protocol.instruct.tool_calls import Tool


class InstructRequest(MistralBase):
    """
    A valid request to be tokenized
    """

    messages: List[ChatMessage]
    system_prompt: Optional[str] = None
    available_tools: Optional[List[Tool]] = None
